extern crate chrono;
extern crate http;
extern crate mime_guess;
extern crate serde;

use chrono::offset::Utc;
use chrono::DateTime;
use http::status::StatusCode;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::io::prelude::*;
use std::net::TcpStream;
use std::path::PathBuf;
use std::time::SystemTime;

#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct ServerConfig {
    pub threads: usize,
    pub port: u16,
    pub bind: String,
    sites: PathBuf,
    routes: PathBuf,
    not_found: PathBuf,
    try_files: Vec<String>,
}

#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct StaticHttpServerRoute {
    domain: String,
    #[serde(default = "empty_string")]
    prefix: String,
    #[serde(default = "empty_path_buf")]
    path: PathBuf,
    #[serde(default = "empty_path_buf")]
    on_404: PathBuf,
}

fn empty_string() -> String {
    "".to_string()
}

fn empty_path_buf() -> PathBuf {
    PathBuf::from("".to_string())
}

#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct StaticHttpServer {
    config: ServerConfig,
    log_accesses: bool,
    routes_date: u128,
    routes: Vec<StaticHttpServerRoute>,
    bound_to: String,
    graph: HashMap<String, Vec<StaticHttpServerRoute>>,
}

impl StaticHttpServer {
    pub fn new(log_accesses: bool, server_config: ServerConfig) -> Self {
        let bt = format!("{}:{}", server_config.bind, server_config.port);
        let mut this = Self {
            log_accesses: log_accesses,
            config: server_config,
            routes_date: 0,
            routes: Vec::new(),
            bound_to: bt,
            graph: HashMap::new(),
        };
        this.check_for_changes();
        this
    }

    pub fn check_for_changes(&mut self) {
        let metadata = match std::fs::metadata(&self.config.routes) {
            Ok(metadata) => metadata,
            Err(_) => {
                std::fs::File::create(&self.config.routes)
                    .unwrap()
                    .write_all(b"")
                    .unwrap();
                std::fs::metadata(&self.config.routes).unwrap()
            }
        };
        let routes_disk_date = metadata
            .modified()
            .unwrap()
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .as_micros();
        if self.routes_date != routes_disk_date {
            self.adapt_from_disk_changes();
        }
    }

    fn adapt_from_disk_changes(&mut self) {
        let mut routes: Vec<StaticHttpServerRoute> = Vec::new();
        std::fs::create_dir_all(self.config.sites.clone()).unwrap();
        for dir_entry_result in self.config.sites.read_dir().unwrap() {
            let dir_entry: std::fs::DirEntry = dir_entry_result.unwrap();
            if !dir_entry.metadata().unwrap().is_dir() {
                eprintln!(
                    "WARNING: {:?} is not a directory! Ignoring...",
                    dir_entry.path()
                );
                continue;
            }
            let site_definition = dir_entry.path().join("site.yaml");
            if !(match site_definition.metadata() {
                Ok(metadata) => metadata.is_file(),
                Err(_) => false,
            }) {
                eprintln!("WARNING: {:?} is not a file! Ignoring...", site_definition);
                continue;
            }
            let site_files = dir_entry.path().join("files");
            if !(match site_files.metadata() {
                Ok(metadata) => metadata.is_dir(),
                Err(_) => false,
            }) {
                eprintln!("WARNING: {:?} is not a directory! Ignoring...", site_files);
                continue;
            }
            let site_definition_string: String = match std::fs::read_to_string(&site_definition) {
                Ok(site_definition_) => site_definition_,
                Err(_) => {
                    eprintln!(
                        "WARNING: {:?} does not have valid text! Ignoring...",
                        site_definition,
                    );
                    continue;
                }
            };
            let mut route_vec: Vec<StaticHttpServerRoute> = match serde_yaml::from_str(
                site_definition_string.as_str(),
            ) {
                Ok(routes_) => routes_,
                Err(_) => {
                    eprintln!(
                            "WARNING: {:?} does not deserialize into a Vec<StaticHttpServerRoutes> struct! Ignoring...",
                            site_definition,
                        );
                    continue;
                }
            };
            for route in route_vec.iter_mut() {
                route.path = site_files.clone();
                let on_404 = route.path.join(route.on_404.clone());
                if on_404.is_file() {
                    route.on_404 = on_404;
                } else {
                    route.on_404 = empty_path_buf();
                }
                if !route.prefix.starts_with('/') {
                    route.prefix = format!("/{}", route.prefix);
                }
                if !route.prefix.ends_with('/') {
                    route.prefix = format!("{}/", route.prefix);
                }
                routes.append(&mut vec![route.clone()]);
            }
        }
        let mut graph: HashMap<String, Vec<StaticHttpServerRoute>> = HashMap::new();
        for route in routes.iter() {
            if !graph.contains_key(&route.domain) {
                graph.insert(route.domain.clone(), Vec::new());
            }
            let graph_vec = graph.get_mut(&route.domain).unwrap();
            graph_vec.append(&mut vec![route.clone()]);
        }
        let mut graph2: HashMap<String, Vec<StaticHttpServerRoute>> = HashMap::new();
        for ge in graph.iter() {
            let mut vector = ge.1.clone();
            vector.sort_by(|x, y| y.prefix.len().cmp(&x.prefix.len()));
            graph2.insert(ge.0.clone(), vector);
        }
        self.graph = graph2;
        std::fs::File::create(&self.config.routes)
            .unwrap()
            .write_all(serde_yaml::to_string(&graph).unwrap().as_bytes())
            .unwrap();
        self.routes_date = std::fs::metadata(&self.config.routes)
            .unwrap()
            .modified()
            .unwrap()
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .as_micros();
    }

    pub fn handle_stream(self, mut stream: TcpStream) {
        let mut buffer: [u8; 2048] = [0; 2048];
        stream.read(&mut buffer).unwrap();
        let request = build_request_from_buffer(&buffer);
        return self.handle_request(&request, &stream);
    }

    fn handle_request(&self, http_request: &HttpRequest, stream: &TcpStream) {
        let host = http_request.headers.get("host").unwrap_or(&self.bound_to);
        let system_time_start = SystemTime::now();
        if self.log_accesses {
            let system_time = system_time_start;
            let datetime: DateTime<Utc> = system_time.into();
            let now_formatted = datetime.format("%Y-%m-%d %T");
            println!(
                "[{}] => ... {} {} request: {}{} by {}",
                now_formatted,
                http_request.http,
                http_request.verb,
                http_request.headers.get("host").unwrap_or(&self.bound_to),
                http_request.path,
                stream.peer_addr().unwrap(),
            );
        }
        let response: HttpResponse = match self.graph.get(host) {
            Some(routes) => {
                let mut response_: HttpResponse = self.generate_404();
                for route in routes.iter() {
                    if http_request.path.starts_with(&route.prefix) {
                        response_ = self.get_for_route(http_request, route);
                        break;
                    }
                }
                response_
            }
            None => self.generate_404(),
        };
        if self.log_accesses {
            let system_time = SystemTime::now();
            let total_time = system_time.duration_since(system_time_start).unwrap();
            let datetime: DateTime<Utc> = system_time.into();
            let now_formatted = datetime.format("%Y-%m-%d %T");
            println!(
                "[{}] <= {:?} {} {} request: {}{} by {}; replied {} bytes on {} microsseconds",
                now_formatted,
                response.http_code,
                http_request.http,
                http_request.verb,
                http_request.headers.get("host").unwrap_or(&self.bound_to),
                http_request.path,
                stream.peer_addr().unwrap(),
                response.payload.len(),
                total_time.as_micros(),
            );
        }
        response.send_reply(stream);
    }

    fn generate_404(&self) -> HttpResponse {
        HttpResponse::new(
            StatusCode::NOT_FOUND,
            std::fs::read(self.config.not_found.clone())
                .unwrap_or(STATIC_404_DEFAULT.clone().as_bytes().to_vec())
                .clone()
                .as_slice(),
            Some("text/html; charset=utf-8".to_string()),
            None,
        )
    }

    fn generate_route_404(&self, route: &StaticHttpServerRoute) -> HttpResponse {
        if route.on_404.to_str().unwrap().len() == 0 {
            self.generate_404()
        } else {
            match std::fs::read(route.on_404.clone()) {
                Ok(contents) => HttpResponse::new(
                    StatusCode::NOT_FOUND,
                    contents.as_slice(),
                    self.content_type_for_exension(&route.on_404.clone().extension()),
                    None,
                ),
                Err(_) => self.generate_404(),
            }
        }
    }

    fn get_for_route(
        &self,
        http_request: &HttpRequest,
        route: &StaticHttpServerRoute,
    ) -> HttpResponse {
        let requested_file = http_request.path[route.prefix.len()..].to_string();
        let route_path = match std::fs::canonicalize(route.path.as_path()) {
            Ok(content) => content,
            Err(_) => {
                return self.generate_route_404(route);
            }
        };
        let requested_relative_path = route_path.join(requested_file);
        let mut possible_paths: Vec<PathBuf> = Vec::new();
        possible_paths.append(&mut vec![requested_relative_path.clone()]);
        for try_file in self.config.try_files.iter() {
            possible_paths.append(&mut vec![requested_relative_path.join(try_file)]);
        }
        for possible_path in possible_paths.iter() {
            match std::fs::canonicalize(possible_path) {
                Ok(canonical_path) => {
                    if canonical_path.starts_with(route_path.clone()) && canonical_path.is_file() {
                        return self.generate_file_response(
                            &canonical_path,
                            route,
                            http_request.headers.get(&"if-modified-since".to_string()),
                        );
                    }
                }
                Err(_) => {}
            }
        }
        self.generate_route_404(route)
    }

    fn generate_file_response(
        &self,
        path: &PathBuf,
        route: &StaticHttpServerRoute,
        modified_since: Option<&String>,
    ) -> HttpResponse {
        let modified_since_parsed: DateTime<Utc> = match modified_since {
            None => DateTime::<Utc>::from(SystemTime::UNIX_EPOCH),
            Some(x) => match DateTime::parse_from_rfc2822(x.as_str()) {
                Ok(x) => x.into(),
                Err(_) => DateTime::<Utc>::from(SystemTime::UNIX_EPOCH),
            },
        };
        match std::fs::read(path) {
            Ok(content) => {
                let browser_cache = modified_since_parsed
                    .signed_duration_since::<Utc>(SystemTime::UNIX_EPOCH.into())
                    .num_seconds();
                let file_time = path
                    .metadata()
                    .unwrap()
                    .modified()
                    .unwrap()
                    .duration_since(SystemTime::UNIX_EPOCH)
                    .unwrap()
                    .as_secs() as i64;
                if browser_cache == file_time {
                    HttpResponse::new(StatusCode::NOT_MODIFIED, vec![].as_slice(), None, None)
                } else {
                    HttpResponse::new(
                        StatusCode::OK,
                        content.clone().as_slice(),
                        self.content_type_for_exension(&path.clone().extension()),
                        Some(path.metadata().unwrap().modified().unwrap()),
                    )
                }
            }
            Err(_) => self.generate_route_404(route),
        }
    }

    fn content_type_for_exension(&self, ext: &Option<&std::ffi::OsStr>) -> Option<String> {
        if ext.is_none() {
            return None;
        }
        let ext2: &str = ext.unwrap().to_str().unwrap();
        let guess: Option<mime_guess::Mime> = mime_guess::from_ext(ext2).first();
        match guess {
            Some(a) => Some(a.essence_str().to_string()),
            _ => None,
        }
    }
}

const STATIC_404_DEFAULT: &str = r###"<!DOCTYPE html>
<html>
    <head>
        <title>Not Found</title>
    </head>
    <body>
        <h1>Not Found</h1>
        <p>The requested resource could not be found.</p>
        <p><a href="https://simple.wikipedia.org/wiki/HTTP_404"><small><em>
            Read mode about this error.
        </em></small></a></p>
    </body>
</html>"###;

pub struct HttpRequest {
    http: String,
    verb: String,
    path: String,
    headers: HashMap<String, String>,
}

pub struct HttpResponse {
    http_code: StatusCode,
    payload: Vec<u8>,
    content_type: Option<String>,
    last_modified: Option<SystemTime>,
}

impl HttpResponse {
    fn new(
        http_code: StatusCode,
        payload: &[u8],
        content_type: Option<String>,
        last_modified: Option<SystemTime>,
    ) -> Self {
        Self {
            http_code: http_code,
            payload: payload.to_vec(),
            content_type: content_type,
            last_modified: last_modified,
        }
    }

    fn send_reply(&self, stream: &TcpStream) {
        reply_http11_response(
            stream,
            &self.http_code,
            self.payload.as_slice(),
            &self.content_type,
            &self.last_modified,
        );
    }
}

fn build_request_from_buffer(buffer: &[u8]) -> HttpRequest {
    let mut first_line_end = match buffer.iter().position(|&s| s == ('\n' as u8)) {
        Some(x) => x,
        None => 0,
    };

    if first_line_end > 0 && buffer[first_line_end - 1] == ('\r' as u8) {
        first_line_end -= 1;
    }

    let first_line = String::from_utf8(
        buffer
            .iter()
            .take(first_line_end)
            .map(|x| *x)
            .collect::<Vec<u8>>(),
    )
    .unwrap();

    let http_verb: String = first_line.split(" ").collect::<Vec<&str>>()[0].to_string();
    let http_version: String = first_line
        .chars()
        .rev()
        .collect::<String>()
        .split(" ")
        .collect::<Vec<&str>>()[0]
        .chars()
        .rev()
        .collect::<String>();
    let http_path_string: String = String::from(
        &first_line[(http_verb.len() + 1)..(first_line.len() - http_version.len() - 1)],
    );
    let http_path = http_path_string;

    let mut header: String = "".to_string();
    let double_new_lines_unix: [u8; 2] = ['\n' as u8, '\n' as u8];
    let double_new_lines_win: [u8; 4] = ['\r' as u8, '\n' as u8, '\r' as u8, '\n' as u8];
    for i in 0..buffer.len() {
        let buffer_part = buffer.split_at(i);
        if buffer_part.1.starts_with(&double_new_lines_unix)
            || buffer_part.1.starts_with(&double_new_lines_win)
        {
            header = std::str::from_utf8(buffer_part.0).unwrap().to_string();
            break;
        }
    }
    let headers: HashMap<String, String> = header
        .lines()
        .skip(1)
        .map(|x: &str| {
            let pair = x.split_at(x.find(':').unwrap());
            (
                pair.0.to_string().to_lowercase(),
                pair.1[1..].trim().to_string(),
            )
        })
        .collect();

    return HttpRequest {
        http: http_version,
        verb: http_verb,
        path: http_path,
        headers: headers,
    };
}

fn reply_http11_response(
    mut stream: &TcpStream,
    http_code: &StatusCode,
    payload: &[u8],
    content_type: &Option<String>,
    last_modified: &Option<SystemTime>,
) {
    stream
        .write(
            format!(
                "HTTP/1.1 {} {}\r\n",
                http_code.as_u16(),
                http_code.canonical_reason().unwrap()
            )
            .as_bytes(),
        )
        .unwrap();
    stream.write("Connection: close\r\n".as_bytes()).unwrap();
    stream
        .write(format!("Content-Length: {}\r\n", payload.len()).as_bytes())
        .unwrap();
    match content_type {
        Some(x) => stream
            .write(format!("Content-Type: {}\r\n", x).as_bytes())
            .unwrap(),
        None => 0usize,
    };
    match last_modified {
        Some(modified_date) => {
            let dt: DateTime<Utc> = modified_date.clone().into();
            let date = dt.format("%a, %d %b %Y %H:%M:%S %z");
            stream
                .write(format!("Last-Modified: {}\r\n", date).as_bytes())
                .unwrap()
        }
        None => 0usize,
    };
    stream.write("\r\n".as_bytes()).unwrap();
    stream.write(payload).unwrap();
    stream.flush().unwrap();
}
